/*
*  Copyright 2018 The Kubernetes Authors.
*  Copyright (c) 2019 NetApp
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*      http://www.apache.org/licenses/LICENSE-2.0
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
 */

package cloud

import (
	"context"
	"fmt"
	"os"
	"regexp"
	"strings"

	"github.com/Azure/azure-sdk-for-go/profiles/latest/compute/mgmt/compute"

	// Other 3rd party packages
	log "github.com/sirupsen/logrus"

	"github.com/Azure/go-autorest/autorest"
	corev1 "k8s.io/api/core/v1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	"github.com/Azure/azure-sdk-for-go/services/network/mgmt/2020-06-01/network"
	"github.com/Azure/go-autorest/autorest/azure/auth"
)

var (
	clientID       string
	clientSecret   string
	subscriptionId string
	tenantID       string
	rgName         string
)

type AzureClient struct {
	authorizer     *autorest.Authorizer
	vnetClient     *network.VirtualNetworksClient
	lbClient       *network.LoadBalancersClient
	vmClient       *compute.VirtualMachineScaleSetsClient
	LoadBalancerIP string
}

func NewAzureClient() (*AzureClient, error) {
	clientID = os.Getenv("AZURE_CLIENT_ID")
	clientSecret = os.Getenv("AZURE_CLIENT_SECRET")
	subscriptionId = os.Getenv("AZURE_SUBSCRIPTION_ID")
	tenantID = os.Getenv("AZURE_TENANT_ID")
	rgName = os.Getenv("AZURE_RESOURCE_GROUP")
	if len(clientID) == 0 {
		return &AzureClient{}, quarkconfig.AzureClientENVNotSetError{Message: "missing AZURE_CLIENT_ID environment variable for Azure client"}
	}
	if len(clientSecret) == 0 {
		return &AzureClient{}, quarkconfig.AzureClientENVNotSetError{Message: "missing AZURE_CLIENT_SECRET environment variable for Azure client"}
	}
	if len(tenantID) == 0 {
		return &AzureClient{}, quarkconfig.AzureClientENVNotSetError{Message: "missing AZURE_TENANT_ID environment variable for Azure client"}
	}
	if len(subscriptionId) == 0 {
		return &AzureClient{}, quarkconfig.AzureClientENVNotSetError{Message: "missing AZURE_SUBSCRIPTION_ID environment variable for Azure client"}
	}
	if len(rgName) == 0 {
		return &AzureClient{}, quarkconfig.AzureClientENVNotSetError{Message: "missing AZURE_RESOURCE_GROUP environment variable for Azure client"}
	}

	config := auth.NewClientCredentialsConfig(clientID, clientSecret, tenantID)
	authorizer, err := config.Authorizer()
	if err != nil {
		return nil, err
	}

	vnetClient := network.NewVirtualNetworksClient(subscriptionId)
	vnetClient.Authorizer = authorizer
	lbClient := network.NewLoadBalancersClient(subscriptionId)
	lbClient.Authorizer = authorizer
	vmClient := compute.NewVirtualMachineScaleSetsClient(subscriptionId)
	vmClient.Authorizer = authorizer

	return &AzureClient{
		authorizer: &authorizer,
		vnetClient: &vnetClient,
		lbClient:   &lbClient,
		vmClient:   &vmClient,
	}, nil
}

func (client *AzureClient) CreateLoadBalancerAddress(_ context.Context, _ string, _ string) error {
	// We can leave this as is until AKS has multi-protocol support and need to share an endpoint between TCP and UDP services.
	return nil
}

func (client *AzureClient) CreateLoadBalancerAddress1P(_ context.Context, _ string, _ []string, _ bool) error {
	// We can leave this as is until AKS has multi-protocol support and need to share an endpoint between TCP and UDP services.
	return nil
}

func (client *AzureClient) DeleteLoadBalancerAddress(_ context.Context, _ string) error {
	return nil
}

func (client *AzureClient) GetLoadBalancerAddress(_ context.Context, _ string) (string, error) {
	// We can leave this as is until AKS will assign an IP
	return "", nil
}

func (client *AzureClient) GetLoadBalancerSubnet(_ context.Context, _ string) (string, error) {
	// We can leave this as is until AKS will assign an IP
	return "", nil
}

func (client *AzureClient) resetInstance(
	ctx context.Context, resourceGroup, vmssName, instanceID string,
) error {
	instanceIDs := []string{instanceID}
	log.WithFields(log.Fields{
		"vmssName":    vmssName,
		"instanceIDs": strings.Join(instanceIDs, ", "),
	}).Debug("Resetting node.")
	_, err := client.vmClient.Restart(
		ctx, resourceGroup, vmssName,
		&compute.VirtualMachineScaleSetVMInstanceIDs{InstanceIds: &instanceIDs})
	if err != nil {
		return fmt.Errorf("failed to reset instance: %s", err.Error())
	}
	return nil
}

// Note that because of how Azure generates resource group names for VM scale
// sets, this will be broken for most clients
func (client *AzureClient) ResetInstance(ctx context.Context, name string) error {
	// TODO: look into a better way to do this without breaking the CloudClient interface?
	// name example: aks-nodepool1-14336892-vmss000001
	re := regexp.MustCompile(`(?P<ScaleSetName>[a-z1-9\-]+-vmss)0+(?P<InstanceID>[0-9a-f]+)`)
	match := re.FindStringSubmatch(name)
	if len(match) < 3 {
		return fmt.Errorf("unable to parse \"%s\"; unexpected format for instance name", name)
	}
	vmssName := match[1]
	instanceID := match[2]
	return client.resetInstance(ctx, rgName, vmssName, instanceID)
}

func (client *AzureClient) ResetInstanceForProviderID(
	ctx context.Context, id string,
) error {
	components := strings.Split(strings.TrimPrefix(id, "azure:///"), "/")
	if len(components) != 10 {
		return fmt.Errorf("provider ID %s not in expected azure:///subscriptions/<subscriber-id>/resourceGroups/<rgName>/providers/Microsoft.Compute/virtualMachineScaleSets/<vmss-id>/virtualMachines/<instanceID> format", id)
	}
	return client.resetInstance(ctx, components[3], components[7], components[9])
}

func (client *AzureClient) ResetInstanceForZone(ctx context.Context, name, _ string) error {
	return client.ResetInstance(ctx, name)
}

// TODO: is this ok?
func (client *AzureClient) CreateManagedInstanceGroup(ctx context.Context, group InstanceGroup) error {
	return nil
}
func (client *AzureClient) GetManagedInstanceGroup(managedInstanceGroupName string) (InstanceGroup, error) {
	return nil, nil
}
func (client *AzureClient) DeleteManagedInstanceGroup(ctx context.Context, managedInstanceGroupName string) error {
	return nil
}
func (client *AzureClient) CreateInstanceForGroup(ctx context.Context, managedGroupName string, numInstancesToCreate int64) error {
	return nil
}
func (client *AzureClient) RecreateInstancesForGroup(ctx context.Context, groupName, zone string, instanceNames ...string) error {
	return nil
}
func (client *AzureClient) ListManagedInstancesForGroup(groupName string) ([]ManagedInstance, error) {
	return nil, nil
}
func (client *AzureClient) DeleteInstanceFromGroup(ctx context.Context, groupName string, instanceNames ...string) error {
	return nil
}

func (client *AzureClient) AddCloudAnnotations(service *corev1.Service, ilbSubnet string, _ bool) {
	service.Annotations[quarkconfig.AzureILBTypeAnnotation] = "true"
	service.Annotations[quarkconfig.AzureILBSubnetAnnotation] = getSubnetNameFromLink(ilbSubnet)
	service.Annotations[quarkconfig.AzureILBHAPortsAnnotation] = "true"
}

func (client *AzureClient) GlobalAccessEnabled(service *corev1.Service) bool {
	log.WithFields(log.Fields{
		"service": service.Name,
	}).Warn("Global access unsupported in Azure.")
	return false
}

func (client *AzureClient) ListZonalNodePools(ctx context.Context) ([]*NodePool, error) {
	return nil, nil
}

func (client *AzureClient) ListRegionalNodePools(ctx context.Context) ([]*NodePool, error) {
	return nil, nil
}

func (client *AzureClient) ResizeRegionalNodePool(ctx context.Context,
	nodePool string, newSize int64) error {
	return nil
}

func (client *AzureClient) IsNotFound(_ error) bool {
	return client.LoadBalancerIP == ""
}
