package v2

import (
	"context"
	"fmt"
	"strings"
	"time"

	log "github.com/sirupsen/logrus"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	utilerrors "k8s.io/apimachinery/pkg/util/errors"
	"k8s.io/client-go/kubernetes"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

// listPods returns all quark deployments resource objects as list
func listPods(k8sclient kubernetes.Interface, namespace string, matchingLabels map[string]string) ([]corev1.Pod, error) {
	// matchingLabels := map[string]string{"app": quarkconfig.QuarkAppName}
	labelSelector := metav1.LabelSelector{MatchLabels: matchingLabels}
	listOptions := metav1.ListOptions{
		LabelSelector: labels.Set(labelSelector.MatchLabels).String(),
	}
	podsList, err := k8sclient.CoreV1().Pods(namespace).List(context.TODO(), listOptions)
	if err != nil {
		return nil, err
	}
	return podsList.Items, nil
}

// listDeployments returns all quark deployments resource objects as list
func listDeployments(k8sclient kubernetes.Interface, namespace string, matchingLabels map[string]string) (*appsv1.DeploymentList, error) {
	// matchingLabels := map[string]string{"app": quarkconfig.QuarkAppName}
	labelSelector := metav1.LabelSelector{MatchLabels: matchingLabels}
	listOptions := metav1.ListOptions{
		LabelSelector: labels.Set(labelSelector.MatchLabels).String(),
	}
	return k8sclient.AppsV1().Deployments(namespace).List(context.TODO(), listOptions)
}

// getDeployment returns k8s deployment object filtered by name in namespace
func getDeployment(k8sclient kubernetes.Interface, namespace, name string) (*appsv1.Deployment, error) {
	depl, err := k8sclient.AppsV1().Deployments(namespace).Get(context.TODO(), name, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	return depl, nil
}

// getStatefulSet returns k8s statefulset object filtered by name in namespace
func getStatefulSet(k8sclient kubernetes.Interface, namespace, name string) (*appsv1.StatefulSet, error) {
	sts, err := k8sclient.AppsV1().StatefulSets(namespace).Get(context.TODO(), name, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	return sts, nil
}

// getDaemonSet returns k8s daemonset object filtered by name in namespace
func getDaemonSet(k8sclient kubernetes.Interface, namespace, name string) (*appsv1.DaemonSet, error) {

	sts, err := k8sclient.AppsV1().DaemonSets(namespace).Get(context.TODO(), name, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	return sts, nil
}

type verifySecretArgs struct {
	k8sclient    kubernetes.Interface
	quarkClient  quarkclientset.Interface
	nvcNamespace string
	// quarkconfig.CertReloadApps* name
	appName string
	// Deployment or StatefulSet
	workloadType string
	// Deployment name or StatefulSet name
	workloadName      string
	namespace         string
	secretName        string
	secretVersion     string
	desiredPodsCount  int32
	verifiedPodsCount int32
}

func getLogHeaderSecretInfo(opts *verifySecretArgs) string {
	return fmt.Sprintf("[%s] [%s/%s]", opts.namespace, opts.secretName, opts.secretVersion)
}

func getLogHeader(opts *verifySecretArgs) string {
	retval := getLogHeaderSecretInfo(opts)
	if opts.workloadName != "" && opts.workloadType != "" {
		retval = fmt.Sprintf("%s %s/%s", retval, strings.ToLower(opts.workloadType), opts.workloadName)
	}
	return retval
}

func verifyPodsForSecretVersion(opts *verifySecretArgs, pods []corev1.Pod, desiredReplicas int32) error {
	if int32(len(pods)) < desiredReplicas {
		return fmt.Errorf("%s has %d pods that is less than desired pods %d count. verification failed",
			getLogHeader(opts), len(pods), desiredReplicas)
	}

	secretNameAnnotationKey := util.GetSecretResourceAnnotationKey(opts.secretName)
	var version string
	var ok bool
	opts.verifiedPodsCount = 0
	for _, pod := range pods {
		if pod.Status.Phase != "Running" {
			continue
		}
		if version, ok = pod.Annotations[secretNameAnnotationKey]; !ok {
			errmsg := fmt.Sprintf("%s pod: %s does not have secret annotation: %s. verification failed",
				getLogHeader(opts), pod.Name, secretNameAnnotationKey)
			log.Debugf(errmsg)
			return fmt.Errorf(errmsg)
		}
		if version != opts.secretVersion {
			errmsg := fmt.Sprintf("%s pod: %s is not updated, secret version annotation (%s) value mismatch. version found: %s",
				getLogHeader(opts), pod.Name, secretNameAnnotationKey, version)
			log.Debugf(errmsg)
			return fmt.Errorf(errmsg)
		}
		opts.verifiedPodsCount++
	}
	if opts.verifiedPodsCount < desiredReplicas {
		return fmt.Errorf("%s not enough pods having latest secret version. %d pods has latest cert but desired pods: %d",
			getLogHeader(opts), opts.verifiedPodsCount, desiredReplicas)
	}
	log.Debugf("%s verification completed; pods has latest certificate. pods verified: %d of out %d",
		getLogHeader(opts), opts.verifiedPodsCount, desiredReplicas)
	return nil
}

func getQuark(quarkClient quarkclientset.Interface, namespace string) (*quarkv1alpha1.Quark, error) {

	quarkList, err := quarkClient.QuarkV1alpha1().Quarks(namespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	if len(quarkList.Items) <= 0 {
		msg := fmt.Sprintf("quark resource not found in namespace %s", namespace)
		log.Errorf(msg)
		return nil, fmt.Errorf(msg)
	}
	return &quarkList.Items[0], err
}

func verifyWorkloadForSecretVersion(opts *verifySecretArgs, matchingLabels map[string]string) error {

	log.Debugf("%s started verification", getLogHeader(opts))

	var workloadDesiredReplicas int32 = 0
	var workloadAvailableReplicas int32 = 0
	var workloadUpdatedReplicas int32 = 0
	var workloadReadyReplicas int32 = 0
	opts.verifiedPodsCount = 0
	opts.desiredPodsCount = 0
	checkReplicas := true
	switch opts.workloadType {
	case "StatefulSet":
		sts, err := getStatefulSet(opts.k8sclient, opts.namespace, opts.workloadName)
		if err != nil {
			return fmt.Errorf("%s error getting statefulset. %v", getLogHeader(opts), err)
		}
		// if available replicas are less than the required replicas, just return error
		if sts.Spec.Replicas != nil {
			workloadDesiredReplicas = *sts.Spec.Replicas
		} else {
			log.Debugf("%s unable to find statefulset spec replicas count, defaulting to 0", getLogHeader(opts))
		}
		workloadUpdatedReplicas = sts.Status.UpdatedReplicas
		workloadReadyReplicas = sts.Status.ReadyReplicas
		workloadAvailableReplicas = sts.Status.AvailableReplicas
	case "Deployment":
		depl, err := getDeployment(opts.k8sclient, opts.namespace, opts.workloadName)
		if err != nil {
			return fmt.Errorf("%s error getting deployment. %v", getLogHeader(opts), err)
		}
		if depl.Spec.Replicas != nil {
			workloadDesiredReplicas = *depl.Spec.Replicas
		} else {
			log.Debugf("%s unable to find deployment spec replicas count, defaulting to 0", getLogHeader(opts))
		}
		workloadUpdatedReplicas = depl.Status.UpdatedReplicas
		workloadReadyReplicas = depl.Status.ReadyReplicas
		workloadAvailableReplicas = depl.Status.AvailableReplicas
	case "DaemonSet":
		dsets, err := getDaemonSet(opts.k8sclient, opts.namespace, opts.workloadName)
		if err != nil {
			return fmt.Errorf("%s error getting daemonsets. %v", getLogHeader(opts), err)
		}
		workloadDesiredReplicas = dsets.Status.DesiredNumberScheduled
		workloadUpdatedReplicas = dsets.Status.UpdatedNumberScheduled
		workloadReadyReplicas = dsets.Status.NumberReady
		workloadAvailableReplicas = dsets.Status.NumberAvailable
		unavailable := dsets.Status.NumberUnavailable
		if unavailable > 0 {
			return fmt.Errorf("%s has %d unavailable pods", getLogHeader(opts), unavailable)
		}
	case "EtcdCluster":
		quark, err := getQuark(opts.quarkClient, opts.nvcNamespace)
		if err != nil {
			return fmt.Errorf("%s error getting quark spec for checking etcd-cluster. %v", getLogHeader(opts), err)
		}
		checkReplicas = false
		workloadDesiredReplicas = int32(quark.Spec.ControllerConfig.EtcdClusterSize)
	default:
		return fmt.Errorf("unknown workload type %s. could not verify certificate update", opts.workloadType)
	}
	opts.desiredPodsCount = workloadDesiredReplicas
	if checkReplicas {
		if workloadDesiredReplicas != workloadUpdatedReplicas {
			return fmt.Errorf("%s desired replicas (%d) and updated replicas(%d) are different",
				getLogHeader(opts), workloadDesiredReplicas, workloadUpdatedReplicas)
		}
		if workloadDesiredReplicas != workloadReadyReplicas {
			return fmt.Errorf("%s desired replicas (%d) and ready replicas(%d) are different",
				getLogHeader(opts), workloadDesiredReplicas, workloadReadyReplicas)
		}
		if workloadReadyReplicas != workloadAvailableReplicas {
			return fmt.Errorf("%s ready replicas (%d) and available replicas(%d) are different",
				getLogHeader(opts), workloadReadyReplicas, workloadAvailableReplicas)
		}
		if workloadDesiredReplicas == 0 {
			log.Debugf("%s has 0 desired replicas. verification skipped", getLogHeader(opts))
			return nil
		}
		if workloadAvailableReplicas < opts.desiredPodsCount {
			return fmt.Errorf("%s has available pods: %d desired pods: %d. verification failed",
				getLogHeader(opts), workloadAvailableReplicas, opts.desiredPodsCount)
		}
	}

	pods, err := listPods(opts.k8sclient, opts.namespace, matchingLabels)
	if err != nil {
		//log.Errorf("unable to get %s pods list %s", opts.appName, err.Error())
		return err
	}

	return verifyPodsForSecretVersion(opts, pods, opts.desiredPodsCount)
}

// verifyNHCAppForSecretVersion - verifies if NHC pods have the secrets updated
func verifyNHCAppForSecretVersion(opts *verifySecretArgs) error {
	opts.workloadName = quarkconfig.NHCStatefulSetName
	opts.workloadType = "StatefulSet"
	matchingLabels := map[string]string{
		"app": opts.workloadName,
	}
	return verifyWorkloadForSecretVersion(opts, matchingLabels)
}

// verifyNVCAppForSecretVersion - verifies if NVC pods have the secrets updated
func verifyNVCAppForSecretVersion(opts *verifySecretArgs) error {
	opts.workloadName = quarkconfig.NVCDeploymentName
	opts.workloadType = "Deployment"
	matchingLabels := map[string]string{
		"app": opts.workloadName,
	}
	return verifyWorkloadForSecretVersion(opts, matchingLabels)
}

func verifyQuarkDeploymentsForSecretVersion(opts *verifySecretArgs) error {
	deplMatchingLabels := map[string]string{"quark.netapp.io/app": quarkconfig.QuarkAppName}
	depls, err := listDeployments(opts.k8sclient, opts.namespace, deplMatchingLabels)
	if err != nil {
		return err
	}
	errs := []error{}
	opts.workloadType = "Deployment"
	verifiedPodsCount := int32(0)
	desiredPodsCount := int32(0)
	for _, depl := range depls.Items {
		desiredPodsCount += depl.Status.Replicas
		opts.workloadName = depl.Name
		matchingLabels := deplMatchingLabels
		matchingLabels["quark.netapp.io/volume"] = depl.Labels["quark.netapp.io/volume"]
		err = verifyWorkloadForSecretVersion(opts, matchingLabels)
		if err != nil {
			errs = append(errs, err)
			continue
		}
		verifiedPodsCount += opts.verifiedPodsCount
	}
	opts.desiredPodsCount = desiredPodsCount
	opts.verifiedPodsCount = verifiedPodsCount

	if len(errs) > 0 {
		return utilerrors.NewAggregate(errs)
	}

	if opts.verifiedPodsCount != opts.desiredPodsCount {
		err = fmt.Errorf("%s storage pool deployments verification failed. expected pods: %d, updated pods: %d",
			getLogHeader(opts), opts.desiredPodsCount, opts.verifiedPodsCount)
		return err
	}

	return nil
}

func verifyEtcdOperatorAppForSecretVersion(opts *verifySecretArgs) error {
	opts.workloadName = quarkconfig.EtcdOperatorDeploymentName
	opts.workloadType = "Deployment"
	matchingLabels := map[string]string{
		"app": opts.workloadName,
	}
	return verifyWorkloadForSecretVersion(opts, matchingLabels)
}

func verifyEtcdClusterAppForSecretVersion(opts *verifySecretArgs) error {
	opts.workloadName = quarkconfig.QuarkEtcdClusterName
	opts.workloadType = "EtcdCluster"
	matchingLabels := map[string]string{
		"app":          "etcd",
		"etcd_cluster": quarkconfig.QuarkEtcdClusterName,
	}
	return verifyWorkloadForSecretVersion(opts, matchingLabels)
}

func verifyNodeMonitorAppForSecretVersion(opts *verifySecretArgs) error {
	opts.workloadName = quarkconfig.NodeMonitorDaemonSetName
	opts.workloadType = "DaemonSet"
	matchingLabels := map[string]string{
		"app": opts.workloadName,
	}
	return verifyWorkloadForSecretVersion(opts, matchingLabels)
}

type certAppLevelUpdateStatus struct {
	status bool // true for application updated with secret, else false
	opts   *verifySecretArgs
}

type applicationsAndSecrets []*certAppLevelUpdateStatus

func getApplicationsForSecrets(nvcClient *NVCClient, secrets []string) (applicationsAndSecrets, error) {
	k8sclient := nvcClient.kubeClient
	quarkClient := nvcClient.quarkClient
	netappVolNamespace := nvcClient.nvcNamespace
	appSecrets := applicationsAndSecrets{}
	util.InitializeCertSecretDataMap(nvcClient.nvcNamespace)
	for _, secretName := range secrets {
		for _, namespacedAppName := range util.GetCertAppDataListForSecret(secretName, "") {
			appNamespace := ""
			appName := ""
			ok := false
			if appNamespace, appName, ok = util.SplitNamespacedName(namespacedAppName); !ok {
				return nil, fmt.Errorf("string (%s) is not a namespaced appname. skipping", namespacedAppName)
			}
			appSecrets = append(appSecrets, &certAppLevelUpdateStatus{
				status: false,
				opts: &verifySecretArgs{
					k8sclient:         k8sclient,
					quarkClient:       quarkClient,
					nvcNamespace:      netappVolNamespace,
					appName:           appName,
					secretName:        secretName,
					secretVersion:     "",
					namespace:         appNamespace,
					workloadType:      "",
					workloadName:      "",
					verifiedPodsCount: 0,
					desiredPodsCount:  0,
				},
			})
		}
	}
	return appSecrets, nil
}

func updateSecretsVersion(nvcClient *NVCClient, updatedAppsAndSecrets *applicationsAndSecrets, secrets []string) error {
	netappVolNamespace := nvcClient.nvcNamespace
	// prefetch secret version. storage format: namespace -> secretName -> secretVersion
	secretsVersion := map[string]map[string]string{
		netappVolNamespace:               {},
		quarkconfig.QuarkSystemNamespace: {},
	}
	// Get latest secret version
	for _, secretName := range secrets {
		if util.IsCertSecretInNetAppVolumeNamespace(secretName) {
			secretVersion, err := util.GetSecretResourceVersion(nvcClient.kubeClient, netappVolNamespace, secretName)
			if err != nil {
				return fmt.Errorf("[%s] [%s] unable to get secret version. error: %v", secretName, netappVolNamespace, err)
			}
			secretsVersion[nvcClient.nvcNamespace][secretName] = secretVersion
		}
		if util.IsCertSecretInQuarkSystemNamespace(secretName) {
			secretVersion, err := util.GetSecretResourceVersion(nvcClient.kubeClient, quarkconfig.QuarkSystemNamespace, secretName)
			if err != nil {
				return fmt.Errorf("[%s] [%s] unable to get secret version. error: %v", secretName, quarkconfig.QuarkSystemNamespace, err)
			}
			secretsVersion[quarkconfig.QuarkSystemNamespace][secretName] = secretVersion
		}
	}
	for _, data := range *updatedAppsAndSecrets {
		if _, ok := secretsVersion[data.opts.namespace]; !ok {
			return fmt.Errorf("%s namespace unrecognized", getLogHeaderSecretInfo(data.opts))
		}
		if _, ok := secretsVersion[data.opts.namespace][data.opts.secretName]; !ok {
			return fmt.Errorf("%s secret unrecognized", getLogHeaderSecretInfo(data.opts))
		}
		newSecretVersion := secretsVersion[data.opts.namespace][data.opts.secretName]
		if data.opts.secretVersion != newSecretVersion {
			data.status = false
			if data.opts.secretVersion != "" {
				data.opts.secretVersion = newSecretVersion
				log.Infof("%s secret changed, re-verification required for %s", getLogHeaderSecretInfo(data.opts), data.opts.appName)
			}
			data.opts.secretVersion = newSecretVersion
		}
	}
	return nil
}

func verifyAppsForCertSecret(updatedAppsAndSecrets *applicationsAndSecrets) error {
	var err error
	for _, data := range *updatedAppsAndSecrets {
		if data.status {
			continue
		}
		err = nil
		opts := data.opts
		logPrinted := false
		switch opts.appName {
		case util.CertReloadAppsNHC:
			err = verifyNHCAppForSecretVersion(opts)
		case util.CertReloadAppsNVC:
			err = verifyNVCAppForSecretVersion(opts)
		case util.CertReloadAppsQuarkDeployments:
			err = verifyQuarkDeploymentsForSecretVersion(opts)
			if err != nil {
				return err
			}
			if opts.desiredPodsCount > 0 {
				log.Infof("%s storage pool deployments has latest certificate (%d pods verified)",
					getLogHeaderSecretInfo(opts), opts.verifiedPodsCount)
			} else {
				log.Infof("%s no storage pool deployments pods available for verification (0 pods verified)", getLogHeaderSecretInfo(opts))
			}
			logPrinted = true
		case util.CertReloadAppsEtcdOperator:
			err = verifyEtcdOperatorAppForSecretVersion(opts)
		case util.CertReloadAppsEtcdCluster:
			err = verifyEtcdClusterAppForSecretVersion(opts)
		case util.CertReloadAppsNodeMonitor:
			err = verifyNodeMonitorAppForSecretVersion(opts)
		default:
			return fmt.Errorf("unknown app for secret version verification: %s", opts.appName)
		}
		if err != nil {
			return err
		}
		if !logPrinted && opts.verifiedPodsCount > 0 {
			log.Infof("%s has latest certificate (%d pods verified)", getLogHeader(opts), opts.verifiedPodsCount)
		}
		data.status = true
	}
	return nil
}

func minOfDuration(first, second time.Duration) time.Duration {
	if first < second {
		return first
	}
	return second
}

func appendAppStatusMsg(prevMsg string, data *certAppLevelUpdateStatus) string {
	if prevMsg != "" {
		prevMsg = prevMsg + ", "
	}
	prevMsg = prevMsg + data.opts.namespace + "/"
	if data.opts.workloadName != "" && data.opts.workloadType != "" {
		prevMsg = prevMsg + strings.ToLower(data.opts.workloadType) + "/" + data.opts.workloadName
	} else {
		prevMsg = prevMsg + data.opts.appName
	}
	return prevMsg
}

// WaitCertificatesReloadComplete is a blocking call that returns when certificates update is completed
func (nvcClient *NVCClient) WaitCertificatesReloadComplete(ctx context.Context, secrets []string, timeout *time.Duration) error {
	defaultTimeout := time.Duration(5) * time.Minute
	if timeout == nil {
		timeout = &defaultTimeout
	}
	for _, secretName := range secrets {
		if !util.IsCertSecretInNetAppVolumeNamespace(secretName) && !util.IsCertSecretInQuarkSystemNamespace(secretName) {
			return fmt.Errorf("invalid secret name to check for certificate update: %s ", secretName)
		}
	}
	// for each secret name, get secret version, then get list of pods to be checked for reload
	// and verify if they match the annotation
	var err error
	var appSecretUpdateStatus applicationsAndSecrets
	appSecretUpdateStatus, err = getApplicationsForSecrets(nvcClient, secrets)
	if err != nil {
		return err
	}
	log.Infof("Waiting for services to be loaded with updated secrets: [ %s ]", strings.Join(secrets, ","))
	intervalTicker := time.NewTicker(minOfDuration(3*time.Second, *timeout))
	timeoutCh := time.After(*timeout)
	for {
		select {
		case <-ctx.Done():
			return fmt.Errorf("context done. exiting")
		case <-intervalTicker.C:
			err := updateSecretsVersion(nvcClient, &appSecretUpdateStatus, secrets)
			if err != nil {
				log.Errorf("error updating secrets: %s", err.Error())
			}
			err = verifyAppsForCertSecret(&appSecretUpdateStatus)
			if err != nil {
				log.Debugf("[waiting] %s. will check again after few seconds", err.Error())
			} else {
				return nil
			}
		case <-timeoutCh:
			completedAppsMsg := ""
			notCompletedAppsMsg := ""
			for _, data := range appSecretUpdateStatus {
				if data.status {
					completedAppsMsg = appendAppStatusMsg(completedAppsMsg, data)
				} else {
					notCompletedAppsMsg = appendAppStatusMsg(notCompletedAppsMsg, data)
				}
			}
			message := ""
			if completedAppsMsg != "" {
				message = "completed-apps: " + completedAppsMsg + "; "
			}
			if notCompletedAppsMsg != "" {
				message = message + "not-completed-apps: " + notCompletedAppsMsg
			}
			return NewNVCClientWaitForTimedoutError(message)
		}
	}
	return nil
}

// WaitTrustedCAsReloadComplete is a blocking call that returns when Trusted CA update is completed
func (nvcClient *NVCClient) WaitTrustedCAsReloadComplete(ctx context.Context, timeout *time.Duration) error {
	secretName := quarkconfig.SecretNameQuarkTrustedCA
	return nvcClient.WaitCertificatesReloadComplete(ctx, []string{secretName}, timeout)
}
