package v2

import (
	"context"
	"fmt"
	"reflect"
	"strings"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

// NetAppVolumeReplicationDelete is used to signal the deletion of a NetAppVolumeReplication.
type NetAppVolumeReplicationDelete struct {
	VolumeReplication *quarkv1alpha1.NetAppVolumeReplication
}

// NetAppVolumeReplicationUpdate is used to signal the most recent state of a NetAppVolumeReplicationUpdate and the
// most recent changes.
type NetAppVolumeReplicationUpdate struct {
	VolumeReplication *quarkv1alpha1.NetAppVolumeReplication
	Status            *NetAppVolumeReplicationStatusInfo
	UpdatedConditions []quarkv1alpha1.NetAppVolumeReplicationCondition
}

// NetAppVolumeReplicationStatusInfo captures the status of any pending updates
type NetAppVolumeReplicationStatusInfo struct {
	MirrorStateUpdated         bool
	ReplicationScheduleUpdated bool
	ReplicationStatusUpdated   bool
}

// ConstructNetAppVolumeReplication creates a NetAppVolumeReplication struct resource
func ConstructNetAppVolumeReplication(
	ownerID, volumeReplicationUUID, storagePoolUUID, clusterPeeringUUID, displayName, destinationVolumeUUID, sourceVolumeUUID string,
	replicationPolicy quarkv1alpha1.NetAppVolumeReplicationPolicyType,
	replicationSchedule quarkv1alpha1.NetAppVolumeReplicationScheduleType,
	mirrorState quarkv1alpha1.NetAppVolumeReplicationMirrorStateType) (volumeReplication *quarkv1alpha1.NetAppVolumeReplication, err error) {
	// validate volumeReplicationUUID
	_, err = uuid.FromString(volumeReplicationUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeReplicationUUID: %v", err)
	}

	_, err = uuid.FromString(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for storagePoolUUID: %v", err)
	}

	_, err = uuid.FromString(clusterPeeringUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for clusterPeeringUUID: %v", err)
	}

	_, err = uuid.FromString(destinationVolumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for destinationVolumeUUID: %v", err)
	}

	_, err = uuid.FromString(sourceVolumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for sourceVolumeUUID: %v", err)
	}

	if len(strings.TrimSpace(ownerID)) == 0 {
		return nil, fmt.Errorf("ownerID cannot be empty string")
	}

	if len(strings.TrimSpace(displayName)) == 0 {
		return nil, fmt.Errorf("display name cannot be empty string")
	}

	if err = util.IsValidReplicationPolicy(replicationPolicy); err != nil {
		return nil, err
	}

	if err = util.IsValidReplicationSchedule(replicationSchedule); err != nil {
		return nil, err
	}

	if mirrorState != quarkv1alpha1.MirrorStateUninitialized &&
		mirrorState != quarkv1alpha1.MirrorStateBroken {
		return nil, fmt.Errorf("invalid mirror state for create: %v. Valid mirror states for create: %v, %v",
			mirrorState, quarkv1alpha1.MirrorStateUninitialized, quarkv1alpha1.MirrorStateBroken)
	}

	volumeReplication = &quarkv1alpha1.NetAppVolumeReplication{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: volumeReplicationUUID,
			Labels: map[string]string{
				quarkconfig.OwnerIDLabel:        ownerID,
				quarkconfig.StoragePoolLabel:    storagePoolUUID,
				quarkconfig.ClusterPeeringLabel: clusterPeeringUUID,
			},
			Finalizers: []string{
				quarkconfig.VolumeReplicationFinalizer,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeReplicationSpec{
			DisplayName: displayName,
			DestinationVolume: quarkv1alpha1.NetAppVolumeReplicationVolume{
				VolumeUUID: destinationVolumeUUID,
			},
			SourceVolume: quarkv1alpha1.NetAppVolumeReplicationVolume{
				VolumeUUID: sourceVolumeUUID,
			},
			MirrorState:         mirrorState,
			ReplicationPolicy:   replicationPolicy,
			ReplicationSchedule: replicationSchedule,
		},
	}
	return
}

// NetAppVolumeReplicationCreate attempts to create a NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationCreate(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) (*quarkv1alpha1.NetAppVolumeReplication, error) {
	storagePoolUUID := volumeReplication.Labels[quarkconfig.StoragePoolLabel]
	_, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting storage pool %v: %v", storagePoolUUID, err)
	}

	clusterPeeringUUID := volumeReplication.Labels[quarkconfig.ClusterPeeringLabel]
	_, err = client.NetAppClusterPeeringGet(clusterPeeringUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting cluster peering %v: %v", clusterPeeringUUID, err)
	}

	dstVolUUID := volumeReplication.Spec.DestinationVolume.VolumeUUID
	dstVol, err := client.NetAppVolumeGet(dstVolUUID)
	if err != nil {
		return nil, fmt.Errorf("error getting destination volume %v: %v", dstVolUUID, err)
	}

	// Check if destination volume belongs to the storagepool supplied in the replication cr
	if dstVol.Labels[quarkconfig.StoragePoolLabel] != storagePoolUUID {
		return nil, fmt.Errorf("destination volume storagepool (%v) not the same as storagepool supplied in cr (%v)", dstVol.Labels[quarkconfig.StoragePoolLabel], storagePoolUUID)
	}

	// dest volume should be dp enabled and not read/write if the volume replication is being created in Uninitialized state
	if volumeReplication.Spec.MirrorState == quarkv1alpha1.MirrorStateUninitialized &&
		dstVol.Spec.DataProtection != quarkv1alpha1.DataProtectionEnabled {
		return nil, fmt.Errorf("volume replication mirror state cannot be set to \"Uninitialized\" since the destination volume %s is not data protection enabled", dstVolUUID)
	}

	// dest volume needs to be read/write if the volume replication is being created in Broken state i.e flip resync scenario
	if volumeReplication.Spec.MirrorState == quarkv1alpha1.MirrorStateBroken &&
		dstVol.Spec.DataProtection == quarkv1alpha1.DataProtectionEnabled {
		return nil, fmt.Errorf("volume replication mirror state cannot be set to \"Broken\" since the destination volume %s has data protection enabled",
			dstVolUUID)
	}

	return client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).Create(
		context.TODO(), volumeReplication, metav1.CreateOptions{})
}

// NetAppVolumeReplicationUpdate attempts to update a NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationUpdate(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) (*quarkv1alpha1.NetAppVolumeReplication, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).Update(
		context.TODO(), volumeReplication, metav1.UpdateOptions{})
}

// NetAppVolumeReplicationDelete attempts to delete a NetAppVolumeReplication CR only if the relationship is broken
func (client *NVCClient) NetAppVolumeReplicationDelete(volumeReplicationUUID string) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if !NetAppVolumeReplicationBroken(volumeReplication) {
		return fmt.Errorf("unable to delete volume replication, volume replication's mirror state must be \"%v\", current state is \"%v\"",
			quarkv1alpha1.MirrorStateBroken,
			volumeReplication.Status.MirrorState)
	}
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).Delete(
		context.TODO(), volumeReplicationUUID, metav1.DeleteOptions{})
}

// NetAppVolumeReplicationGet attempts to retrieve a NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationGet(volumeReplicationUUID string) (*quarkv1alpha1.NetAppVolumeReplication, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).Get(
		context.TODO(), volumeReplicationUUID, metav1.GetOptions{})
}

// NetAppVolumeReplicationGetAll attempts to retrieve all the NetAppVolumeReplication CRs
func (client *NVCClient) NetAppVolumeReplicationGetAll() ([]*quarkv1alpha1.NetAppVolumeReplication, error) {
	volumeReplicationsList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	volumeReplications := make([]*quarkv1alpha1.NetAppVolumeReplication, len(volumeReplicationsList.Items))
	for i := range volumeReplicationsList.Items {
		volumeReplications[i] = &volumeReplicationsList.Items[i]
	}
	return volumeReplications, err
}

// NetAppVolumeReplicationMirrorState returns the MirrorState of the NetAppVolumeReplication resource
func NetAppVolumeReplicationMirrorState(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) quarkv1alpha1.NetAppVolumeReplicationMirrorStateType {
	return volumeReplication.Status.MirrorState
}

// NetAppVolumeReplicationMirrored checks if the Replication relationship is Mirrored
func NetAppVolumeReplicationMirrored(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateMirrored)
}

// NetAppVolumeReplicationBroken checks if the Replication relationship is Broken
func NetAppVolumeReplicationBroken(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return (volumeReplication.Status.MirrorState == quarkv1alpha1.MirrorStateBroken)
}

// NetAppVolumeReplicationStatus returns the MirrorState of the NetAppVolumeReplication resource
// Expected values are one of ReplicationStatusIdle, ReplicationStatusTransferring
func NetAppVolumeReplicationStatus(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) quarkv1alpha1.NetAppVolumeReplicationStatusType {
	return volumeReplication.Status.ReplicationStatus
}

// NetAppVolumeReplicationScheduleUpdated checks if the Spec and the Status ReplicationSchedule field match
// A match would indicate a successful update
func NetAppVolumeReplicationScheduleUpdated(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	return volumeReplication.Spec.ReplicationSchedule == volumeReplication.Status.ReplicationSchedule
}

// NetAppVolumeReplicationHealthy checks if the relationship is healthy
func NetAppVolumeReplicationHealthy(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	for _, cond := range volumeReplication.Status.Conditions {
		if cond.Type == quarkv1alpha1.VolumeReplicationHealthy && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppVolumeReplicationTransferSnapshotError checks if the current transfer of the snapshot failed
func NetAppVolumeReplicationTransferSnapshotError(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) error {
	for _, cond := range volumeReplication.Status.Conditions {
		if cond.Type == quarkv1alpha1.VolumeReplicationTransferError && cond.Status == corev1.ConditionTrue {
			return fmt.Errorf(cond.Message)
		}
	}
	return nil
}

// NetAppVolumeReplicationTransferSnapshotAborted checks if a given snapshot transfer is aborted by checking the appropriate condition
func NetAppVolumeReplicationTransferSnapshotAborted(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) bool {
	for _, cond := range volumeReplication.Status.Conditions {
		if cond.Type == quarkv1alpha1.VolumeReplicationTransferAborted && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppVolumeReplicationInitialize is used to start the initialize operation
// Initialization is done automatically, but can also be driven by the CVS API in case it initially failed.
func (client *NVCClient) NetAppVolumeReplicationInitialize(volumeReplicationUUID, transferSnapshot string) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if volumeReplication.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored &&
		volumeReplication.Spec.TransferSnapshot == transferSnapshot {
		return nil
	}
	if volumeReplication.Status.MirrorState != quarkv1alpha1.MirrorStateUninitialized {
		return fmt.Errorf("volume replication state must be \"%v\". Current state is \"%v\"",
			quarkv1alpha1.MirrorStateUninitialized, volumeReplication.Status.MirrorState)
	}

	volumeReplicationCopy := volumeReplication.DeepCopy()
	volumeReplicationCopy.Spec.MirrorState = quarkv1alpha1.MirrorStateMirrored
	volumeReplicationCopy.Spec.TransferSnapshot = transferSnapshot
	_, err = client.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

// NetAppVolumeReplicationResync is used to start a resync operation
// Resync is performed when the user wants to resume mirroring data after a Break.
// Resync can only be done on a Broken relationship.
func (client *NVCClient) NetAppVolumeReplicationResync(volumeReplicationUUID string) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if volumeReplication.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored {
		return nil
	}
	if !NetAppVolumeReplicationBroken(volumeReplication) {
		return fmt.Errorf("volume replication must be %v", quarkv1alpha1.MirrorStateBroken)
	}

	// Set volume's Spec.dataProtection field to true for resync
	err = client.netAppVolumeDataProtectionUpdate(
		volumeReplication.Spec.DestinationVolume.VolumeUUID,
		quarkv1alpha1.DataProtectionEnabled)
	if err != nil {
		return fmt.Errorf(
			"unable to update dataProtection for destination volume %v, error %v",
			volumeReplication.Spec.DestinationVolume.VolumeUUID, err.Error())
	}

	volumeReplicationCopy := volumeReplication.DeepCopy()
	volumeReplicationCopy.Spec.MirrorState = quarkv1alpha1.MirrorStateMirrored
	_, err = client.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	if err != nil {
		// Set dataProtection field to false on the volume
		// Ignore any errors since the operation to resync would likely be retried
		if volErr := client.netAppVolumeDataProtectionUpdate(
			volumeReplication.Spec.DestinationVolume.VolumeUUID,
			quarkv1alpha1.DataProtectionDisabled); volErr != nil {
			log.WithFields(log.Fields{
				"volume":            volumeReplication.Spec.DestinationVolume.VolumeUUID,
				"volumeReplication": volumeReplication.Name,
				"dataProtection":    quarkv1alpha1.MirrorStateMirrored,
				"error":             volErr.Error(),
			}).Warn("Unable to update dataProtection for volume.")
		}
	}
	return err
}

// NetAppVolumeReplicationBreak is used to start the break operation
// Break is performed when the user wants to read/write to the secondary volume.
// This might be done, for example, if the primary region is down. Break can only be done on a Mirrored relationship.
func (client *NVCClient) NetAppVolumeReplicationBreak(volumeReplicationUUID string) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if volumeReplication.Spec.MirrorState == quarkv1alpha1.MirrorStateBroken {
		return nil
	}
	if !NetAppVolumeReplicationMirrored(volumeReplication) {
		return fmt.Errorf("volume replication state must be \"%v\". Current state is \"%v\"",
			quarkv1alpha1.MirrorStateMirrored, volumeReplication.Status.MirrorState)
	}

	// Set volume's Spec.dataProtection field to false
	err = client.netAppVolumeDataProtectionUpdate(
		volumeReplication.Spec.DestinationVolume.VolumeUUID,
		quarkv1alpha1.DataProtectionDisabled)
	if err != nil {
		return fmt.Errorf(
			"unable to update dataProtection for destination volume %v, error %v",
			volumeReplication.Spec.DestinationVolume.VolumeUUID, err.Error())
	}

	volumeReplicationCopy := volumeReplication.DeepCopy()
	volumeReplicationCopy.Spec.MirrorState = quarkv1alpha1.MirrorStateBroken
	_, err = client.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	if err != nil {
		// Set dataProtection field to true on the volume
		// Ignore any errors since the break operation would likely be retried
		if volErr := client.netAppVolumeDataProtectionUpdate(
			volumeReplication.Spec.DestinationVolume.VolumeUUID,
			quarkv1alpha1.DataProtectionEnabled); volErr != nil {
			log.WithFields(log.Fields{
				"volume":            volumeReplication.Spec.DestinationVolume.VolumeUUID,
				"volumeReplication": volumeReplication.Name,
				"error":             volErr.Error(),
				"dataProtection":    quarkv1alpha1.MirrorStateBroken,
			}).Warn("Unable to update dataProtection for volume.")
		}
	}
	return err
}

// NetAppVolumeReplicationTransferSnapshot is used to start the transfer of a snapshot identified by name.
// Transfer Snapshot is initiated internally by the Replication Scheduler in SDE.
// It will be performed every 10 minutes, hourly, or daily based on the schedule chosen by the customer.
// NOTE: Transfer snapshot is not supported for the first release of CRR.
func (client *NVCClient) NetAppVolumeReplicationTransferSnapshot(volumeReplicationUUID, snapshotName string) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if volumeReplication.Spec.TransferSnapshot == snapshotName {
		return nil
	}
	if !NetAppVolumeReplicationMirrored(volumeReplication) {
		return fmt.Errorf("volume replication state must be \"%v\". Current state is \"%v\"",
			quarkv1alpha1.MirrorStateMirrored, volumeReplication.Status.MirrorState)
	}
	// Disallow new transfer if one is in progress - consumer should first abort
	if volumeReplication.Spec.TransferSnapshot != volumeReplication.Status.TransferSnapshot ||
		volumeReplication.Status.ReplicationStatus == quarkv1alpha1.ReplicationStatusTransferring ||
		(NetAppVolumeReplicationTransferSnapshotError(volumeReplication) != nil &&
			!NetAppVolumeReplicationTransferSnapshotAborted(volumeReplication)) {
		return fmt.Errorf("cannot transfer while another transfer or abort is in progress")
	}
	volumeReplicationCopy := volumeReplication.DeepCopy()
	volumeReplicationCopy.Spec.TransferSnapshot = snapshotName
	_, err = client.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

// NetAppVolumeReplicationTransferAbort is used to abort a transfer that is started by NetAppVolumeReplicationTransferSnapshot()
// NOTE: Aborting transfer is not supported for the first release
func (client *NVCClient) NetAppVolumeReplicationTransferAbort(volumeReplicationUUID string) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if volumeReplication.Spec.TransferSnapshot == "" {
		return nil
	}
	volumeReplicationCopy := volumeReplication.DeepCopy()
	volumeReplicationCopy.Spec.TransferSnapshot = ""
	_, err = client.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

// NetAppVolumeReplicationScheduleModify is used to modify the ReplicationSchedule
// Accepted values are ReplicationSchedule10Minutely, ReplicationScheduleHourly and ReplicationScheduleDaily
func (client *NVCClient) NetAppVolumeReplicationScheduleModify(volumeReplicationUUID string, schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	volumeReplication, err := client.NetAppVolumeReplicationGet(volumeReplicationUUID)
	if err != nil {
		return err
	}
	if volumeReplication.Spec.ReplicationSchedule == schedule {
		return nil
	}
	volumeReplicationCopy := volumeReplication.DeepCopy()
	volumeReplicationCopy.Spec.ReplicationSchedule = schedule
	_, err = client.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

// NetAppVolumeReplicationUpdated checks for specific modifications/ updates to the NetAppVolumeReplication CR
func (client *NVCClient) NetAppVolumeReplicationUpdated(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) *NetAppVolumeReplicationStatusInfo {
	volumeReplicationUpdate := &NetAppVolumeReplicationStatusInfo{
		MirrorStateUpdated:         true,
		ReplicationScheduleUpdated: true,
	}

	if volumeReplication.Spec.MirrorState != volumeReplication.Status.MirrorState {
		volumeReplicationUpdate.MirrorStateUpdated = false
	}
	if volumeReplication.Spec.ReplicationSchedule != volumeReplication.Status.ReplicationSchedule {
		volumeReplicationUpdate.ReplicationScheduleUpdated = false
	}
	log.WithFields(log.Fields{
		"MirrorStateUpdated":         volumeReplicationUpdate.MirrorStateUpdated,
		"ReplicationScheduleUpdated": volumeReplicationUpdate.ReplicationScheduleUpdated,
	}).Debug("NetAppVolumeReplication state has been updated.")
	return volumeReplicationUpdate
}

// handleNetappVolumeReplicationUpdate is used to publish NetAppVolumeReplication specific events
// This is done by comparing the previous and current state of the NetAppVolumeReplication CR
func (client *NVCClient) handleNetappVolumeReplicationUpdate(
	oldVolumeReplication *quarkv1alpha1.NetAppVolumeReplication,
	newVolumeReplication *quarkv1alpha1.NetAppVolumeReplication) {
	sendEvent := false
	newVolumeReplicationUpdate := client.NetAppVolumeReplicationUpdated(newVolumeReplication)
	oldVolumeReplicationUpdate := client.NetAppVolumeReplicationUpdated(oldVolumeReplication)
	if newVolumeReplicationUpdate != oldVolumeReplicationUpdate {
		log.WithFields(log.Fields{
			"volumeReplication": newVolumeReplication.Name,
			"oldStatus":         oldVolumeReplicationUpdate,
			"newStatus":         newVolumeReplicationUpdate,
		}).Debug("NetApp Volume Replication state has been updated.")
		sendEvent = true
	}
	// We send events when the NetAppVolumeReplications's ReplicationStatus field has changed.
	if newVolumeReplication.Status.ReplicationStatus != oldVolumeReplication.Status.ReplicationStatus {
		sendEvent = true
		newVolumeReplicationUpdate.ReplicationStatusUpdated = true
		log.WithFields(log.Fields{
			"volumeReplication": newVolumeReplication.Name,
			"oldStatus":         oldVolumeReplicationUpdate,
			"newStatus":         newVolumeReplicationUpdate,
		}).Debug("NetApp Volume Replication has updated ReplicationStatus.")
	}
	// We send events when the NetAppVolumeReplications's conditions have changed.
	if !reflect.DeepEqual(oldVolumeReplication.Status.Conditions, newVolumeReplication.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"volumeReplication": newVolumeReplication.Name,
			"conditions":        newVolumeReplication.Status.Conditions,
		}).Debug("NetApp Volume Replication has updated conditions.")

	}
	if sendEvent {
		client.volumeReplicationUpdateChannel <- &NetAppVolumeReplicationUpdate{
			VolumeReplication: newVolumeReplication,
			Status:            newVolumeReplicationUpdate,
			UpdatedConditions: newVolumeReplication.Status.Conditions,
		}
	}
}

// handleVolumeReplicationDelete is used to terminate the watch on VolumeReplication events
func (client *NVCClient) handleVolumeReplicationDelete(volumeReplication *quarkv1alpha1.NetAppVolumeReplication) {
	client.volumeReplicationDeleteChannel <- &NetAppVolumeReplicationDelete{VolumeReplication: volumeReplication}
}
